import os
from setuptools import setup, find_packages

# Note: setup() has access to cmd arguments of the setup.py script via sys.argv

# Utility function to read the README file.
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


install_requires = [
    "pickle5",
    "matplotlib",
    "tqdm",
    "tensorboard",
    "torch>=1.8",
    "torch-geometric>=1.7.2",
    "torch-scatter",
    "torch-sparse",
]
test_requires = ["black", "pytest", "pytest-mpi"]

setup(
    name="HydraGNN",
    version="3.0",
    package_dir={"hydragnn": "hydragnn"},
    packages=find_packages(),
    install_requires=install_requires,
    extras_require={"test": test_requires},
    description="Distributed PyTorch implementation of multi-headed graph convolutional neural networks",
    license="BSD-3",
    long_description_content_type="text/markdown",
    long_description=read("README.md"),
    url="https://github.com/ORNL/HydraGNN",
    author="Massimiliano Lupo Pasini, Samuel Temple Reeve, Pei Zhang, Jong Youl Choi",
    author_email="lupopasinim@ornl.gov",
)



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import os
import numpy as np

import torch
import torch.distributed as dist
from torch.utils.tensorboard import SummaryWriter
from torch_geometric.data import Data
from torch_geometric.utils import degree
from .print_utils import print_master, iterate_tqdm

from hydragnn.utils.distributed import (
    get_comm_size_and_rank,
    get_device,
    get_device_name,
)
from collections import OrderedDict


def activation_function_selection(activation_function_string: str):
    if activation_function_string == "relu":
        return torch.nn.ReLU()
    elif activation_function_string == "selu":
        return torch.nn.SELU()
    elif activation_function_string == "prelu":
        return torch.nn.PReLU()
    elif activation_function_string == "elu":
        return torch.nn.ELU()
    elif activation_function_string == "lrelu_01":
        return torch.nn.LeakyReLU(0.1)
    elif activation_function_string == "lrelu_025":
        return torch.nn.LeakyReLU(0.25)
    elif activation_function_string == "lrelu_05":
        return torch.nn.LeakyReLU(0.5)


def loss_function_selection(loss_function_string: str):
    if loss_function_string == "mse":
        return torch.nn.functional.mse_loss
    elif loss_function_string == "mae":
        return torch.nn.functional.l1_loss
    elif loss_function_string == "smooth_l1":
        return torch.nn.SmoothL1Loss
    elif loss_function_string == "rmse":
        return lambda x, y: torch.sqrt(torch.nn.functional.mse_loss(x, y))


def save_model(model, optimizer, name, path="./logs/"):
    """Save both model and optimizer state in a single checkpoint file"""
    _, world_rank = get_comm_size_and_rank()
    if hasattr(optimizer, "consolidate_state_dict"):
        optimizer.consolidate_state_dict()
    if world_rank == 0:
        path_name = os.path.join(path, name, name + ".pk")
        torch.save(
            {
                "model_state_dict": model.state_dict(),
                "optimizer_state_dict": optimizer.state_dict(),
            },
            path_name,
        )


def get_summary_writer(name, path="./logs/"):
    _, world_rank = get_comm_size_and_rank()
    if world_rank == 0:
        path_name = os.path.join(path, name)
        writer = SummaryWriter(path_name)


def load_existing_model_config(model, config, path="./logs/", optimizer=None):
    if "continue" in config and config["continue"]:
        model_name = config["startfrom"]
        load_existing_model(model, model_name, path, optimizer)


def load_existing_model(model, model_name, path="./logs/", optimizer=None):
    """Load both model and optimizer state from a single checkpoint file"""
    path_name = os.path.join(path, model_name, model_name + ".pk")
    map_location = {"cuda:%d" % 0: get_device_name()}
    print_master("Load existing model:", path_name)
    checkpoint = torch.load(path_name, map_location=map_location)
    state_dict = checkpoint["model_state_dict"]
    ## To be compatible with old checkpoint which was not written as a ddp model
    if not next(iter(state_dict)).startswith("module"):
        ddp_state_dict = OrderedDict()
        for k, v in state_dict.items():
            k = "module." + k
            ddp_state_dict[k] = v
        state_dict = ddp_state_dict
    model.load_state_dict(state_dict)
    if (optimizer is not None) and ("optimizer_state_dict" in checkpoint):
        optimizer.load_state_dict(checkpoint["optimizer_state_dict"])


## This function may cause OOM if dataset is too large
## to fit in a single GPU (i.e., with DDP). Use with caution.
## Recommend to use calculate_PNA_degree_dist
def calculate_PNA_degree(loader, max_neighbours):
    backend = os.getenv("HYDRAGNN_AGGR_BACKEND", "torch")
    if backend == "torch":
        return calculate_PNA_degree_dist(loader, max_neighbours)
    elif backend == "mpi":
        return calculate_PNA_degree_mpi(loader, max_neighbours)
    else:
        deg = torch.zeros(max_neighbours + 1, dtype=torch.long)
        for data in iterate_tqdm(loader, 2, desc="Calculate PNA degree"):
            d = degree(data.edge_index[1], num_nodes=data.num_nodes, dtype=torch.long)
            deg += torch.bincount(d, minlength=deg.numel())[: max_neighbours + 1]
        return deg


def calculate_PNA_degree_dist(loader, max_neighbours):
    assert dist.is_initialized()
    deg = torch.zeros(max_neighbours + 1, dtype=torch.long)
    for data in iterate_tqdm(loader, 2, desc="Calculate PNA degree"):
        d = degree(data.edge_index[1], num_nodes=data.num_nodes, dtype=torch.long)
        deg += torch.bincount(d, minlength=deg.numel())[: max_neighbours + 1]
    deg = deg.to(get_device())
    dist.all_reduce(deg, op=dist.ReduceOp.SUM)
    deg = deg.detach().cpu()
    return deg


def calculate_PNA_degree_mpi(loader, max_neighbours):
    assert dist.is_initialized()
    deg = torch.zeros(max_neighbours + 1, dtype=torch.long)
    for data in iterate_tqdm(loader, 2, desc="Calculate PNA degree"):
        d = degree(data.edge_index[1], num_nodes=data.num_nodes, dtype=torch.long)
        deg += torch.bincount(d, minlength=deg.numel())[: max_neighbours + 1]
    from mpi4py import MPI

    deg = MPI.COMM_WORLD.allreduce(deg.numpy(), op=MPI.SUM)
    return torch.tensor(deg)


def unsorted_segment_mean(data, segment_ids, num_segments):
    result_shape = (num_segments, data.size(1))
    segment_ids = segment_ids.unsqueeze(-1).expand(-1, data.size(1))
    result = data.new_full(result_shape, 0)  # Init empty result tensor.
    count = data.new_full(result_shape, 0)
    result.scatter_add_(0, segment_ids, data)
    count.scatter_add_(0, segment_ids, torch.ones_like(data))
    return result / count.clamp(min=1)


def print_model(model):
    """print model's parameter size layer by layer"""
    num_params = 0
    for k, v in model.state_dict().items():
        print_master("%50s\t%20s\t%10d" % (k, list(v.shape), v.numel()))
        num_params += v.numel()
    print_master("-" * 50)
    print_master("%50s\t%20s\t%10d" % ("Total", "", num_params))
    print_master("All (total, MB): %d %g" % (num_params, num_params * 4 / 1024 / 1024))


def tensor_divide(x1, x2):
    return torch.from_numpy(np.divide(x1, x2, out=np.zeros_like(x1), where=x2 != 0))


# early stop based on validation loss
class EarlyStopping:
    def __init__(self, patience=10, min_delta=0.0):
        self.patience = patience
        self.min_delta = min_delta
        self.val_loss_min = float("inf")
        self.count = 0

    def __call__(self, val_loss):
        if val_loss > self.val_loss_min + self.min_delta:
            self.count += 1
            if self.count >= self.patience:
                return True
        else:
            self.val_loss_min = val_loss
            self.count = 0
        return False


class Checkpoint:
    """
    Checkpoints the model and optimizer when:
        + The performance metric is smaller than the stored performance metric
    Args
      warmup: (int) Number of epochs to warmup prior to checkpointing.
      path: (str) Path for checkpointing
      name: (str) Model name for the directory and the file to save.
    """

    def __init__(
        self,
        name: str,
        warmup: int = 0,
        path: str = "./logs/",
    ):
        self.count = 1
        self.warmup = warmup
        self.path = path
        self.name = name
        self.min_perf_metric = float("inf")
        self.min_delta = 0

    def __call__(self, model, optimizer, perf_metric):

        if (perf_metric > self.min_perf_metric + self.min_delta) or (
            self.count < self.warmup
        ):
            self.count += 1
            return False
        else:
            self.min_perf_metric = perf_metric
            save_model(model, optimizer, name=self.name, path=self.path)
            return True



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import pickle
from sklearn.model_selection import StratifiedShuffleSplit

import torch
from torch_geometric.data import Data
from torch_geometric.transforms import (
    Distance,
    NormalizeRotation,
    Spherical,
    PointPairFeatures,
)

from hydragnn.preprocess import update_predicted_values, update_atom_features
from hydragnn.utils.distributed import get_device
from hydragnn.utils.print_utils import print_distributed, iterate_tqdm
from hydragnn.preprocess.utils import (
    get_radius_graph,
    get_radius_graph_pbc,
)


class SerializedDataLoader:
    """A class used for loading existing structures from files that are lists of serialized structures.
    Most of the class methods are hidden, because from outside a caller needs only to know about
    load_serialized_data method.
    """

    """
    Constructor
    """

    def __init__(self, config, dist=False):
        self.verbosity = config["Verbosity"]["level"]
        self.node_feature_name = config["Dataset"]["node_features"]["name"]
        self.node_feature_dim = config["Dataset"]["node_features"]["dim"]
        self.node_feature_col = config["Dataset"]["node_features"]["column_index"]
        self.graph_feature_name = config["Dataset"]["graph_features"]["name"]
        self.graph_feature_dim = config["Dataset"]["graph_features"]["dim"]
        self.graph_feature_col = config["Dataset"]["graph_features"]["column_index"]
        self.rotational_invariance = config["Dataset"]["rotational_invariance"]
        self.periodic_boundary_conditions = config["NeuralNetwork"]["Architecture"][
            "periodic_boundary_conditions"
        ]
        self.radius = config["NeuralNetwork"]["Architecture"]["radius"]
        self.max_neighbours = config["NeuralNetwork"]["Architecture"]["max_neighbours"]
        self.variables = config["NeuralNetwork"]["Variables_of_interest"]
        self.variables_type = config["NeuralNetwork"]["Variables_of_interest"]["type"]
        self.output_index = config["NeuralNetwork"]["Variables_of_interest"][
            "output_index"
        ]
        self.input_node_features = config["NeuralNetwork"]["Variables_of_interest"][
            "input_node_features"
        ]

        self.spherical_coordinates = False
        self.point_pair_features = False

        if "Descriptors" in config["Dataset"]:
            if "SphericalCoordinates" in config["Dataset"]["Descriptors"]:
                self.spherical_coordinates = config["Dataset"]["Descriptors"][
                    "SphericalCoordinates"
                ]
            if "PointPairFeatures" in config["Dataset"]["Descriptors"]:
                self.point_pair_features = config["Dataset"]["Descriptors"][
                    "PointPairFeatures"
                ]

        self.subsample_percentage = None

        # In situations where someone already provides the .pkl filed with data
        # the asserts from raw_dataset_loader are not performed
        # Therefore, we need to re-check consistency
        assert len(self.node_feature_name) == len(self.node_feature_dim)
        assert len(self.node_feature_name) == len(self.node_feature_col)
        assert len(self.graph_feature_name) == len(self.graph_feature_dim)
        assert len(self.graph_feature_name) == len(self.graph_feature_col)

        self.dist = dist
        if self.dist:
            assert torch.distributed.is_initialized()
            self.world_size = torch.distributed.get_world_size()
            self.rank = torch.distributed.get_rank()

    """
    Methods
    -------
    load_serialized_data(dataset_path: str, config: dict)
        Loads the serialized structures data from specified path, computes new edges for the structures based on the maximum number of neighbours and radius. Additionally,
        atom and structure features are updated.
    """

    def load_serialized_data(self, dataset_path: str):
        """Loads the serialized structures data from specified path, computes new edges for the structures based on the maximum number of neighbours and radius. Additionally,
        atom and structure features are updated.

        Parameters
        ----------
        dataset_path: str
            Directory path where files containing serialized structures are stored.
        config: dict
            Dictionary containing information needed to load the data and transform it, respectively: atom_features, radius, max_num_node_neighbours and predicted_value_option.
        Returns
        ----------
        [Data]
            List of Data objects representing atom structures.
        """
        with open(dataset_path, "rb") as f:
            _ = pickle.load(f)
            _ = pickle.load(f)
            dataset = pickle.load(f)

        rotational_invariance = NormalizeRotation(max_points=-1, sort=False)
        if self.rotational_invariance:
            dataset[:] = [rotational_invariance(data) for data in dataset]

        if self.periodic_boundary_conditions:
            # edge lengths already added manually if using PBC, so no need to call Distance.
            compute_edges = get_radius_graph_pbc(
                radius=self.radius,
                loop=False,
                max_neighbours=self.max_neighbours,
            )
        else:
            compute_edges = get_radius_graph(
                radius=self.radius,
                loop=False,
                max_neighbours=self.max_neighbours,
            )

        dataset[:] = [compute_edges(data) for data in dataset]

        # edge lengths already added manually if using PBC.
        if not self.periodic_boundary_conditions:
            compute_edge_lengths = Distance(norm=False, cat=True)
            dataset[:] = [compute_edge_lengths(data) for data in dataset]

        max_edge_length = torch.Tensor([float("-inf")])

        for data in dataset:
            max_edge_length = torch.max(max_edge_length, torch.max(data.edge_attr))

        if self.dist:
            ## Gather max in parallel
            device = max_edge_length.device
            max_edge_length = max_edge_length.to(get_device())
            torch.distributed.all_reduce(
                max_edge_length, op=torch.distributed.ReduceOp.MAX
            )
            max_edge_length = max_edge_length.to(device)

        # Normalization of the edges
        for data in dataset:
            data.edge_attr = data.edge_attr / max_edge_length

        # Descriptors about topology of the local environment
        if self.spherical_coordinates:
            self.dataset[:] = [Spherical(data) for data in self.dataset]

        if self.point_pair_features:
            self.dataset[:] = [PointPairFeatures(data) for data in self.dataset]

        # Move data to the device, if used. # FIXME: this does not respect the choice set by use_gpu
        device = get_device(verbosity_level=self.verbosity)
        for data in dataset:
            ## (2022/04) jyc: no need for parallel loading
            # data.to(device)
            update_predicted_values(
                self.variables_type,
                self.output_index,
                self.graph_feature_dim,
                self.node_feature_dim,
                data,
            )

            update_atom_features(self.input_node_features, data)

        if "subsample_percentage" in self.variables.keys():
            self.subsample_percentage = self.variables["subsample_percentage"]
            return self.__stratified_sampling(
                dataset=dataset, subsample_percentage=self.subsample_percentage
            )

        return dataset

    def __stratified_sampling(self, dataset: [Data], subsample_percentage: float):
        """Given the dataset and the percentage of data you want to extract from it, method will
        apply stratified sampling where X is the dataset and Y is are the category values for each datapoint.
        In the case of the structures dataset where each structure contains 2 types of atoms, the category will
        be constructed in a way: number of atoms of type 1 + number of protons of type 2 * 100.

        Parameters
        ----------
        dataset: [Data]
            A list of Data objects representing a structure that has atoms.
        subsample_percentage: float
            Percentage of the dataset.

        Returns
        ----------
        [Data]
            Subsample of the original dataset constructed using stratified sampling.
        """
        dataset_categories = []
        print_distributed(
            self.verbosity, "Computing the categories for the whole dataset."
        )
        for data in iterate_tqdm(dataset, self.verbosity):
            frequencies = torch.bincount(data.x[:, 0].int())
            frequencies = sorted(frequencies[frequencies > 0].tolist())
            category = 0
            for index, frequency in enumerate(frequencies):
                category += frequency * (100 ** index)
            dataset_categories.append(category)

        subsample_indices = []
        subsample = []

        sss = StratifiedShuffleSplit(
            n_splits=1, train_size=subsample_percentage, random_state=0
        )

        for subsample_index, rest_of_data_index in sss.split(
            dataset, dataset_categories
        ):
            subsample_indices = subsample_index.tolist()

        for index in subsample_indices:
            subsample.append(dataset[index])

        return subsample



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import os, json
from functools import singledispatch

import torch
import torch.distributed as dist
from torch.optim.lr_scheduler import ReduceLROnPlateau

from hydragnn.preprocess.load_data import dataset_loading_and_splitting
from hydragnn.utils.distributed import (
    setup_ddp,
    get_distributed_model,
    print_peak_memory,
)
from hydragnn.utils.model import (
    save_model,
    get_summary_writer,
    load_existing_model_config,
)
from hydragnn.utils.print_utils import print_distributed, setup_log
from hydragnn.utils.time_utils import print_timers
from hydragnn.utils.config_utils import (
    update_config,
    get_log_name_config,
    save_config,
)
from hydragnn.utils.optimizer import select_optimizer
from hydragnn.models.create import create_model_config
from hydragnn.train.train_validate_test import train_validate_test


@singledispatch
def run_training(config):
    raise TypeError("Input must be filename string or configuration dictionary.")


@run_training.register
def _(config_file: str):

    with open(config_file, "r") as f:
        config = json.load(f)

    run_training(config)


@run_training.register
def _(config: dict):

    try:
        os.environ["SERIALIZED_DATA_PATH"]
    except:
        os.environ["SERIALIZED_DATA_PATH"] = os.getcwd()

    setup_log(get_log_name_config(config))
    world_size, world_rank = setup_ddp()

    train_loader, val_loader, test_loader = dataset_loading_and_splitting(config=config)

    config = update_config(config, train_loader, val_loader, test_loader)
    plot_init_solution = config["Visualization"]["plot_init_solution"]
    plot_hist_solution = config["Visualization"]["plot_hist_solution"]
    create_plots = config["Visualization"]["create_plots"]

    model = create_model_config(
        config=config["NeuralNetwork"], verbosity=config["Verbosity"]["level"]
    )
    print_peak_memory(
        config["Verbosity"]["level"], "Max memory allocated after creating local model"
    )
    model = get_distributed_model(
        model,
        config["Verbosity"]["level"],
        sync_batch_norm=config["NeuralNetwork"]["Architecture"]["SyncBatchNorm"],
    )
    print_peak_memory(
        config["Verbosity"]["level"],
        "Max memory allocated after creating distributed model",
    )

    optimizer = select_optimizer(
        model, config["NeuralNetwork"]["Training"]["Optimizer"]
    )

    scheduler = ReduceLROnPlateau(
        optimizer, mode="min", factor=0.5, patience=5, min_lr=0.00001
    )

    log_name = get_log_name_config(config)
    writer = get_summary_writer(log_name)

    if dist.is_initialized():
        dist.barrier()

    save_config(config, log_name)

    load_existing_model_config(
        model, config["NeuralNetwork"]["Training"], optimizer=optimizer
    )

    print_distributed(
        config["Verbosity"]["level"],
        f"Starting training with the configuration: \n{json.dumps(config, indent=4, sort_keys=True)}",
    )

    train_validate_test(
        model,
        optimizer,
        train_loader,
        val_loader,
        test_loader,
        writer,
        scheduler,
        config["NeuralNetwork"],
        log_name,
        config["Verbosity"]["level"],
        plot_init_solution,
        plot_hist_solution,
        create_plots,
    )

    save_model(model, optimizer, log_name)

    print_timers(config["Verbosity"]["level"])



##############################################################################
# Copyright (c) 2021, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of HydraGNN and is distributed under a BSD 3-clause      #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# SPDX-License-Identifier: BSD-3-Clause                                      #
##############################################################################

import torch
from torch_geometric.transforms import RadiusGraph
from torch_geometric.utils import remove_self_loops, degree
from torch_geometric.data import Data

import ase
import ase.neighborlist
import os

from .dataset_descriptors import AtomFeatures

## This function can be slow if dataset is too large. Use with caution.
## Recommend to use check_if_graph_size_variable_dist
def check_if_graph_size_variable(train_loader, val_loader, test_loader):
    backend = os.getenv("HYDRAGNN_AGGR_BACKEND", "torch")
    if backend == "torch":
        return check_if_graph_size_variable_dist(train_loader, val_loader, test_loader)
    elif backend == "mpi":
        return check_if_graph_size_variable_mpi(train_loader, val_loader, test_loader)
    else:
        graph_size_variable = False
        nodes_num_list = []
        for loader in [train_loader, val_loader, test_loader]:
            for data in loader.dataset:
                nodes_num_list.append(data.num_nodes)
            if len(list(set(nodes_num_list))) > 1:
                graph_size_variable = True
                return graph_size_variable
        return graph_size_variable


def check_if_graph_size_variable_dist(train_loader, val_loader, test_loader):
    from hydragnn.utils.distributed import get_device

    assert torch.distributed.is_initialized()
    graph_size_variable = False
    nodes_num_list = []
    for loader in [train_loader, val_loader, test_loader]:
        for data in loader:
            nodes_num_list.append(data.num_nodes)
            if len(list(set(nodes_num_list))) > 1:
                graph_size_variable = True
                break
        if graph_size_variable:
            break
    b = 1 if graph_size_variable else 0
    b = torch.tensor(b).to(get_device())
    torch.distributed.all_reduce(b, op=torch.distributed.ReduceOp.SUM)
    reduced = True if b.item() > 0 else False
    return reduced


def check_if_graph_size_variable_mpi(train_loader, val_loader, test_loader):
    graph_size_variable = False
    nodes_num_list = []
    for loader in [train_loader, val_loader, test_loader]:
        for data in loader:
            nodes_num_list.append(data.num_nodes)
            if len(list(set(nodes_num_list))) > 1:
                graph_size_variable = True
                break
        if graph_size_variable:
            break
    from mpi4py import MPI

    b = 1 if graph_size_variable else 0
    rb = MPI.COMM_WORLD.allreduce(b, op=MPI.SUM)
    reduced = True if rb > 0 else False
    return reduced


def check_data_samples_equivalence(data1, data2, tol):
    x_bool = data1.x.shape == data2.x.shape
    pos_bool = data1.pos.shape == data2.pos.shape
    y_bool = data1.y.shape == data2.y.shape

    found = [False for i in range(data1.edge_index.shape[1])]
    for i in range(data1.edge_index.shape[1]):
        for j in range(data1.edge_index.shape[1]):
            if torch.equal(data1.edge_index[:, i], data2.edge_index[:, j]):
                found[i] = True
                # Due to numerics, the assert check fails for discrepancies below 1e-6
                assert torch.norm(data1.edge_attr[i] - data2.edge_attr[j]) < tol
                break

    edge_bool = all(found)

    return x_bool and pos_bool and y_bool and edge_bool


def get_radius_graph(radius, max_neighbours, loop=False):
    return RadiusGraph(
        r=radius,
        loop=loop,
        max_num_neighbors=max_neighbours,
    )


def get_radius_graph_pbc(radius, max_neighbours, loop=False):
    return RadiusGraphPBC(
        r=radius,
        loop=loop,
        max_num_neighbors=max_neighbours,
    )


def get_radius_graph_config(config, loop=False):
    return RadiusGraph(
        r=config["radius"],
        loop=loop,
        max_num_neighbors=config["max_neighbours"],
    )


def get_radius_graph_pbc_config(config, loop=False):
    return RadiusGraphPBC(
        r=config["radius"],
        loop=loop,
        max_num_neighbors=config["max_neighbours"],
    )


class RadiusGraphPBC(RadiusGraph):
    r"""Creates edges based on node positions :obj:`pos` to all points within a
    given distance, including periodic images.
    """

    def __call__(self, data):
        data.edge_attr = None
        assert (
            "batch" not in data
        ), "Periodic boundary conditions not currently supported on batches."
        assert hasattr(
            data, "supercell_size"
        ), "The data must contain the size of the supercell to apply periodic boundary conditions."
        ase_atom_object = ase.Atoms(
            positions=data.pos,
            cell=data.supercell_size,
            pbc=True,
        )
        # ‘i’ : first atom index
        # ‘j’ : second atom index
        # https://wiki.fysik.dtu.dk/ase/ase/neighborlist.html#ase.neighborlist.neighbor_list
        edge_src, edge_dst, edge_length = ase.neighborlist.neighbor_list(
            "ijd", a=ase_atom_object, cutoff=self.r, self_interaction=self.loop
        )
        data.edge_index = torch.stack(
            [torch.LongTensor(edge_src), torch.LongTensor(edge_dst)], dim=0
        )

        # ensure no duplicate edges
        num_edges = data.edge_index.size(1)
        data.coalesce()
        assert num_edges == data.edge_index.size(
            1
        ), "Adding periodic boundary conditions would result in duplicate edges. Cutoff radius must be reduced or system size increased."

        data.edge_attr = torch.tensor(edge_length, dtype=torch.float).unsqueeze(1)

        return data

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(r={self.r})"


def gather_deg(dataset):
    from hydragnn.utils.print_utils import iterate_tqdm

    backend = os.getenv("HYDRAGNN_AGGR_BACKEND", "torch")
    if backend == "torch":
        return gather_deg_dist(dataset)
    elif backend == "mpi":
        return gather_deg_mpi(dataset)
    else:
        max_deg = 0
        for data in iterate_tqdm(dataset, 2, desc="Degree max"):
            d = degree(data.edge_index[1], num_nodes=data.num_nodes, dtype=torch.long)
            max_deg = max(max_deg, max(d))

        deg = torch.zeros(max_deg + 1, dtype=torch.long)
        for data in iterate_tqdm(dataset, 2, desc="Degree bincount"):
            d = degree(data.edge_index[1], num_nodes=data.num_nodes, dtype=torch.long)
            deg += torch.bincount(d, minlength=deg.numel())
        return deg


def gather_deg_dist(dataset):
    import torch.distributed as dist
    from hydragnn.utils.print_utils import iterate_tqdm
    from hydragnn.utils.distributed import get_device

    max_deg = 0
    for data in iterate_tqdm(dataset, 2, desc="Degree max"):
        d = degree(data.edge_index[1], num_nodes=data.num_nodes, dtype=torch.long)
        max_deg = max(max_deg, int(d.max()))
    max_deg = torch.tensor(max_deg, requires_grad=False).to(get_device())
    dist.all_reduce(max_deg, op=dist.ReduceOp.MAX)

    deg = torch.zeros(max_deg.item() + 1, requires_grad=False, dtype=torch.long)
    for data in iterate_tqdm(dataset, 2, desc="Degree bincount"):
        d = degree(data.edge_index[1], num_nodes=data.num_nodes, dtype=torch.long)
        deg += torch.bincount(d, minlength=deg.numel())
    deg = deg.clone().detach().to(get_device())
    dist.all_reduce(deg, op=dist.ReduceOp.SUM)
    return deg.cpu().detach().numpy()


def gather_deg_mpi(dataset):
    from mpi4py import MPI
    from hydragnn.utils.print_utils import iterate_tqdm

    max_deg = 0
    for data in iterate_tqdm(dataset, 2, desc="Degree max"):
        d = degree(data.edge_index[1], num_nodes=data.num_nodes, dtype=torch.long)
        max_deg = max(max_deg, max(d))
    max_deg = MPI.COMM_WORLD.allreduce(max_deg, op=MPI.MAX)

    deg = torch.zeros(max_deg + 1, dtype=torch.long)
    for data in iterate_tqdm(dataset, 2, desc="Degree bincount"):
        d = degree(data.edge_index[1], num_nodes=data.num_nodes, dtype=torch.long)
        deg += torch.bincount(d, minlength=deg.numel())
    deg = MPI.COMM_WORLD.allreduce(deg.numpy(), op=MPI.SUM)
    return deg


def update_predicted_values(
    type: list, index: list, graph_feature_dim: list, node_feature_dim: list, data: Data
):
    """Updates values of the structure we want to predict. Predicted value is represented by integer value.
    Parameters
    ----------
    type: "graph" level or "node" level
    index: index/location in data.y for graph level and in data.x for node level
    graph_feature_dim: list of integers to trak the dimension of each graph level feature
    data: Data
        A Data object representing a structure that has atoms.
    """
    output_feature = []
    data.y_loc = torch.zeros(1, len(type) + 1, dtype=torch.int64, device=data.x.device)
    for item in range(len(type)):
        if type[item] == "graph":
            index_counter_global_y = sum(graph_feature_dim[: index[item]])
            feat_ = torch.reshape(
                data.y[
                    index_counter_global_y : index_counter_global_y
                    + graph_feature_dim[index[item]]
                ],
                (graph_feature_dim[index[item]], 1),
            )
            # after the global features are spanned, we need to iterate over the nodal features
            # to do so, the counter of the nodal features need to start from the last value of counter for the graph nodel feature
        elif type[item] == "node":
            index_counter_nodal_y = sum(node_feature_dim[: index[item]])
            feat_ = torch.reshape(
                data.x[
                    :,
                    index_counter_nodal_y : (
                        index_counter_nodal_y + node_feature_dim[index[item]]
                    ),
                ],
                (-1, 1),
            )
        else:
            raise ValueError("Unknown output type", type[item])
        output_feature.append(feat_)
        data.y_loc[0, item + 1] = data.y_loc[0, item] + feat_.shape[0] * feat_.shape[1]
    data.y = torch.cat(output_feature, 0)


def update_atom_features(atom_features: [AtomFeatures], data: Data):
    """Updates atom features of a structure. An atom is represented with x,y,z coordinates and associated features.

    Parameters
    ----------
    atom_features: [AtomFeatures]
        List of features to update. Each feature is instance of Enum AtomFeatures.
    data: Data
        A Data object representing a structure that has atoms.
    """
    feature_indices = [i for i in atom_features]
    data.x = data.x[:, feature_indices]


def stratified_sampling(dataset: [Data], subsample_percentage: float, verbosity=0):
    """Given the dataset and the percentage of data you want to extract from it, method will
    apply stratified sampling where X is the dataset and Y is are the category values for each datapoint.
    In the case of the structures dataset where each structure contains 2 types of atoms, the category will
    be constructed in a way: number of atoms of type 1 + number of protons of type 2 * 100.

    Parameters
    ----------
    dataset: [Data]
        A list of Data objects representing a structure that has atoms.
    subsample_percentage: float
        Percentage of the dataset.

    Returns
    ----------
    [Data]
        Subsample of the original dataset constructed using stratified sampling.
    """
    dataset_categories = []
    print_distributed(verbosity, "Computing the categories for the whole dataset.")
    for data in iterate_tqdm(dataset, verbosity):
        frequencies = torch.bincount(data.x[:, 0].int())
        frequencies = sorted(frequencies[frequencies > 0].tolist())
        category = 0
        for index, frequency in enumerate(frequencies):
            category += frequency * (100 ** index)
        dataset_categories.append(category)

    subsample_indices = []
    subsample = []

    sss = StratifiedShuffleSplit(
        n_splits=1, train_size=subsample_percentage, random_state=0
    )

    for subsample_index, rest_of_data_index in sss.split(dataset, dataset_categories):
        subsample_indices = subsample_index.tolist()

    for index in subsample_indices:
        subsample.append(dataset[index])

    return subsample
